import pytest

from xchainpy2_thorchain.memo import THORMemo, AUTO_OPTIMIZED, ActionType, Affiliate


def test_invalid_memo():
    assert THORMemo.parse_memo('foo:bar:123', no_raise=True) is None
    with pytest.raises(NotImplementedError):
        assert THORMemo.parse_memo('baobab:palm')


def test_noop():
    n1 = THORMemo.noop()
    assert n1.action == ActionType.NOOP
    assert not n1.no_vault
    assert n1.build() == 'NOOP'

    n2 = THORMemo.noop(True)
    assert n2.action == ActionType.NOOP
    assert n2.no_vault
    assert n2.build() == 'NOOP:NOVAULT'

    assert THORMemo.parse_memo('noop') == n1
    assert THORMemo.parse_memo('NOOP') == n1
    assert THORMemo.parse_memo('nOoP') == n1
    assert THORMemo.parse_memo('nOoP:abra') == n1

    assert THORMemo.parse_memo('nOoP:novault') == n2
    assert THORMemo.parse_memo('NOOP:NOVAULT') == n2
    assert THORMemo.parse_memo('Noop:NoVault') == n2


def test_reserve():
    memo = THORMemo.reserve()
    assert memo.build() == 'RESERVE'
    assert memo.action == ActionType.RESERVE

    assert THORMemo.parse_memo('reserve') == memo


TX_ID = '12345679123456789AABB434434DDE11ADA'


def test_out():
    o = THORMemo.outbound(TX_ID)
    assert o.action == ActionType.OUTBOUND
    assert o.tx_id == TX_ID

    assert o.build() == f'OUT:{TX_ID}'

    assert THORMemo.parse_memo(f'OUT:{TX_ID}') == o
    assert THORMemo.parse_memo('OUT:123') == THORMemo.outbound('123')
    assert THORMemo.parse_memo('OUT:') == THORMemo.outbound('')


def test_refund():
    o = THORMemo.refund(TX_ID)
    assert o.action == ActionType.REFUND
    assert o.tx_id == TX_ID

    assert o.build() == f'REFUND:{TX_ID}'

    assert THORMemo.parse_memo(f'REFUND:{TX_ID}') == o
    assert THORMemo.parse_memo('REFUND:123') == THORMemo.refund('123')
    assert THORMemo.parse_memo('REFUND:') == THORMemo.refund('')


THOR_ADDR_1 = 'thor1dl7un46w7l7f3ewrnrm6nq58nerjtp0dradjtd'
THOR_ADDR_2 = 'thor104gsqwta048e80j909g6y9kkqdjrw0lff866ew'


def test_leave():
    memo = THORMemo.leave(THOR_ADDR_1)

    assert memo.action == ActionType.LEAVE
    assert memo.node_address == THOR_ADDR_1

    assert memo.build() == f'LEAVE:{THOR_ADDR_1}'

    assert THORMemo.parse_memo(f'LEAVE:{THOR_ADDR_1}') == memo


def test_unbond():
    u = THORMemo.unbond(THOR_ADDR_1, 1234567890, THOR_ADDR_2)

    assert u.action == ActionType.UNBOND
    assert u.node_address == THOR_ADDR_1
    assert u.provider_address == THOR_ADDR_2
    assert u.amount == 1234567890

    assert u.build() == f'UNBOND:{THOR_ADDR_1}:1234567890:{THOR_ADDR_2}'
    assert THORMemo.parse_memo(f'unbond:{THOR_ADDR_1}:1234567890:{THOR_ADDR_2}') == u

    u2 = THORMemo.unbond(THOR_ADDR_1, 777)
    assert u2.action == ActionType.UNBOND
    assert u2.node_address == THOR_ADDR_1
    assert u2.provider_address == ''
    assert u2.build() == f'UNBOND:{THOR_ADDR_1}:777'

    assert THORMemo.parse_memo('UNBOND:foo:5:bar') == THORMemo.unbond('foo', 5, 'bar')

    # todo: should it raise an exception?
    assert THORMemo.parse_memo('UNBOND:foo:non_number:bar') == THORMemo.unbond('foo', 0, 'bar')


def text_bond():
    u = THORMemo.bond(THOR_ADDR_1, THOR_ADDR_2, 200)

    assert u.action == ActionType.UNBOND
    assert u.node_address == THOR_ADDR_1
    assert u.provider_address == THOR_ADDR_2
    assert u.affiliate_fee_bp == 200

    assert u.build() == f'BOND:{THOR_ADDR_1}:{THOR_ADDR_2}:200'
    assert THORMemo.parse_memo(f'bond:{THOR_ADDR_1}:{THOR_ADDR_2}:200') == u

    # no fee
    u = THORMemo.bond(THOR_ADDR_1, THOR_ADDR_2)
    assert u.build() == f'BOND:{THOR_ADDR_1}:{THOR_ADDR_2}'
    assert THORMemo.parse_memo(f'bond:{THOR_ADDR_1}:{THOR_ADDR_2}') == u

    # todo: should it raise?
    assert THORMemo.parse_memo(f'bond:{THOR_ADDR_1}:{THOR_ADDR_2}:foo') == u

    # just node
    u = THORMemo.bond(THOR_ADDR_1)
    assert u.build() == f'BOND:{THOR_ADDR_1}'
    assert THORMemo.parse_memo(f'BOND:{THOR_ADDR_1}') == u


ETH_ADDR = '0x1c7b17362c84287bd1184447e6dfeaf920c31bbe'
ETH_ADDR_2 = '0xe9973cb51ee04446a54ffca73446d33f133d2f49'
BTC_ADDR = 'bc1qp2t4hl4jr6wjfzv28tsdyjysw7p5armf7px55w'


def test_loan_close():
    m = THORMemo.loan_close('ETH.ETH', ETH_ADDR, 1000)
    assert m.action == ActionType.LOAN_CLOSE
    assert m.asset == 'ETH.ETH'
    assert m.dest_address == ETH_ADDR

    assert m.build() == f'$-:ETH.ETH:{ETH_ADDR}:1000'

    assert THORMemo.parse_memo(f'Loan-:ETH.ETH:{ETH_ADDR}:1000') == m
    assert THORMemo.parse_memo(f'$-:ETH.ETH:{ETH_ADDR}:1000') == m

    m = THORMemo.loan_close('ETH.ETH', ETH_ADDR)
    assert m.action == ActionType.LOAN_CLOSE
    assert m.asset == 'ETH.ETH'
    assert m.dest_address == ETH_ADDR
    assert m.build() == f'$-:ETH.ETH:{ETH_ADDR}'

    assert THORMemo.parse_memo(f'$-:BTC.BTC:{BTC_ADDR}') == \
           THORMemo.loan_close('BTC.BTC', BTC_ADDR)

    assert THORMemo.parse_memo(f'LOAN-:ETH.ETH:{ETH_ADDR_2}:404204059') == \
           THORMemo.loan_close('ETH.ETH', ETH_ADDR_2, 404204059)

    assert THORMemo.parse_memo(f'LOAN-:ETH.ETH:{ETH_ADDR_2}:404e8') == \
           THORMemo.loan_close('ETH.ETH', ETH_ADDR_2, 40400000000)


def test_loan_open():
    m = THORMemo.loan_open('BTC.BTC', BTC_ADDR, 282392832)
    assert m.action == ActionType.LOAN_OPEN
    assert m.asset == 'BTC.BTC'
    assert m.dest_address == BTC_ADDR
    assert m.limit == 282392832
    assert m.affiliate_fee_bp == 0
    assert m.affiliate_address == ''

    assert m.build() == f'$+:BTC.BTC:{BTC_ADDR}:282392832'
    assert THORMemo.parse_memo(f'$+:BTC.BTC:{BTC_ADDR}:282392832') == m

    m = THORMemo.loan_open('ETH.ETH', ETH_ADDR, 404204059, 't', 30)
    assert m.action == ActionType.LOAN_OPEN
    assert m.asset == 'ETH.ETH'
    assert m.dest_address == ETH_ADDR
    assert m.limit == 404204059
    assert m.affiliate_fee_bp == 30
    assert m.affiliate_address == 't'
    assert m.dex_aggregator_address == ''
    assert m.min_amount_out == 0
    assert m.final_asset_address == ''

    assert m.build() == f'$+:ETH.ETH:{ETH_ADDR}:404204059:t:30'

    assert THORMemo.parse_memo(f'Loan+:ETH.ETH:{ETH_ADDR}:404204059:t:30') == m
    assert THORMemo.parse_memo(f'$+:ETH.ETH:{ETH_ADDR}:404204059:t:30') == m

    m = THORMemo.loan_open('BTC.BTC', BTC_ADDR, 334343434, affiliates=[
        Affiliate('t', 30), Affiliate('dx', 40), Affiliate(THOR_ADDR_1, 15)
    ])
    assert m.action == ActionType.LOAN_OPEN
    assert m.asset == 'BTC.BTC'
    assert m.dest_address == BTC_ADDR
    assert m.limit == 334343434
    with pytest.raises(ValueError):
        assert m.affiliate_fee_bp
    with pytest.raises(ValueError):
        assert m.affiliate_address

    assert m.affiliates == [
        Affiliate('t', 30), Affiliate('dx', 40), Affiliate(THOR_ADDR_1, 15)
    ]
    assert m.build() == f'$+:BTC.BTC:{BTC_ADDR}:334343434:t/dx/{THOR_ADDR_1}:30/40/15'


FINAL_ETH_ASSET = '0xA0B86991C6218B36C1D19D4A2E9EB0CE3606EB48'
USDC = 'ETH.USDC-0XA0B86991C6218B36C1D19D4A2E9EB0CE3606EB48'


def test_thor_name():
    m = THORMemo.thorname_register_or_renew('foo', 'THOR', THOR_ADDR_1)
    assert m.action == ActionType.THORNAME
    assert m.name == 'foo'
    assert m.chain == 'THOR'
    assert m.dest_address == THOR_ADDR_1
    assert m.build() == f'~:foo:THOR:{THOR_ADDR_1}'
    assert THORMemo.parse_memo(f'~:foo:THOR:{THOR_ADDR_1}') == m

    m = THORMemo.thorname_register_or_renew('acc-test', 'THOR', THOR_ADDR_1, THOR_ADDR_2)
    assert m.action == ActionType.THORNAME
    assert m.name == 'acc-test'
    assert m.chain == 'THOR'
    assert m.dest_address == THOR_ADDR_1
    assert m.owner == THOR_ADDR_2
    assert m.build() == f'~:acc-test:THOR:{THOR_ADDR_1}:{THOR_ADDR_2}'
    assert THORMemo.parse_memo(f'N:acc-test:THOR:{THOR_ADDR_1}:{THOR_ADDR_2}') == m

    m = THORMemo.thorname_register_or_renew('some_name', 'THOR', THOR_ADDR_1, THOR_ADDR_2, USDC)
    assert m.action == ActionType.THORNAME
    assert m.name == 'some_name'
    assert m.chain == 'THOR'
    assert m.dest_address == THOR_ADDR_1
    assert m.owner == THOR_ADDR_2
    assert m.affiliate_asset == USDC
    assert m.build() == f'~:some_name:THOR:{THOR_ADDR_1}:{THOR_ADDR_2}:{USDC}'
    assert THORMemo.parse_memo(f'~:some_name:THOR:{THOR_ADDR_1}:{THOR_ADDR_2}:{USDC}') == m
    assert THORMemo.parse_memo(f'name:some_name:THOR:{THOR_ADDR_1}:{THOR_ADDR_2}:{USDC}') == m


def test_donate():
    m = THORMemo.donate(USDC)
    assert m.action == ActionType.DONATE
    assert m.asset == USDC
    assert m.build() == f'DONATE:{USDC}'
    assert THORMemo.parse_memo(f'DONATE:{USDC}') == m
    assert THORMemo.parse_memo(f'donate:{USDC}') == m
    assert THORMemo.parse_memo(f'd:{USDC}') == m


def test_withdraw():
    m = THORMemo.withdraw(USDC, 10000)
    assert m.action == ActionType.WITHDRAW
    assert m.pool == USDC
    assert m.withdraw_portion_bp == 10000
    assert m.build() == f'-:{USDC}:10000'
    assert THORMemo.parse_memo(f'-:{USDC}:10000') == m

    m = THORMemo.withdraw('BTC.BTC')
    assert m.action == ActionType.WITHDRAW
    assert m.pool == 'BTC.BTC'
    assert m.withdraw_portion_bp == 10000  # default
    assert m.build() == f'-:BTC.BTC:10000'
    assert THORMemo.parse_memo(f'-:BTC.BTC') == m

    m = THORMemo.withdraw('BTC.BTC', 5000, 'RUNE')
    assert m == THORMemo.withdraw_rune('BTC.BTC', 5000)
    assert m.action == ActionType.WITHDRAW
    assert m.pool == 'BTC.BTC'
    assert m.withdraw_portion_bp == 5000
    assert m.asset == 'RUNE'
    assert m.build() == f'-:BTC.BTC:5000:RUNE'
    assert THORMemo.parse_memo(f'-:BTC.BTC:5000:RUNE') == m

    m = THORMemo.withdraw('BTC.BTC', 5000, 'BTC.BTC')
    assert m.action == ActionType.WITHDRAW
    assert m.pool == 'BTC.BTC'
    assert m.withdraw_portion_bp == 5000
    assert m.asset == 'BTC.BTC'
    assert m.build() == f'-:BTC.BTC:5000:BTC.BTC'
    assert THORMemo.parse_memo(f'-:BTC.BTC:5000:BTC.BTC') == m


def test_swap():
    m = THORMemo.swap('ETH.ETH', ETH_ADDR)
    assert m.action == ActionType.SWAP
    assert m.pool == 'ETH.ETH'
    assert m.dest_address == ETH_ADDR
    assert m.refund_address == m.dest_address
    assert m.build() == f'=:ETH.ETH:{ETH_ADDR}'
    assert THORMemo.parse_memo(f'=:ETH.ETH:{ETH_ADDR}') == m

    m = THORMemo.swap('THOR.RUNE', THOR_ADDR_1, 9999999)
    assert m.action == ActionType.SWAP
    assert m.pool == 'THOR.RUNE'
    assert m.dest_address == THOR_ADDR_1
    assert m.refund_address == m.dest_address
    assert m.limit == 9999999
    assert m.build() == f'=:THOR.RUNE:{THOR_ADDR_1}:9999999'
    assert THORMemo.parse_memo(f'=:THOR.RUNE:{THOR_ADDR_1}:9999999') == m

    m = THORMemo.swap('THOR.RUNE', THOR_ADDR_1, 5555, 33, 99)
    assert m.action == ActionType.SWAP
    assert m.pool == 'THOR.RUNE'
    assert m.dest_address == THOR_ADDR_1
    assert m.refund_address == m.dest_address
    assert m.limit == 5555
    assert m.s_swap_interval == 33
    assert m.s_swap_quantity == 99
    assert m.build() == f'=:THOR.RUNE:{THOR_ADDR_1}:5555/33/99'
    assert THORMemo.parse_memo(f'=:THOR.RUNE:{THOR_ADDR_1}:5555/33/99') == m

    m = THORMemo.swap('THOR.RUNE', THOR_ADDR_1, 5555, 33,
                      s_swap_quantity=AUTO_OPTIMIZED,
                      affiliate_address='dx', affiliate_fee_bp=35)
    assert m.action == ActionType.SWAP
    assert m.pool == 'THOR.RUNE'
    assert m.dest_address == THOR_ADDR_1
    assert m.refund_address == m.dest_address
    assert m.limit == 5555
    assert m.s_swap_interval == 33
    assert m.s_swap_quantity == AUTO_OPTIMIZED
    assert m.affiliate_address == 'dx'
    assert m.affiliate_fee_bp == 35
    assert m.build() == f'=:THOR.RUNE:{THOR_ADDR_1}:5555/33/0:dx:35'
    assert THORMemo.parse_memo(f'=:THOR.RUNE:{THOR_ADDR_1}:5555/33/0:dx:35') == m

    m = THORMemo.swap('ETH.ETH', ETH_ADDR, dex_aggregator_address=ETH_ADDR_2,
                      dex_final_asset_address=FINAL_ETH_ASSET, dex_min_amount_out=333888)
    assert m.action == ActionType.SWAP
    assert m.pool == 'ETH.ETH'
    assert m.dest_address == ETH_ADDR
    assert m.refund_address == m.dest_address
    assert m.dex_aggregator_address == ETH_ADDR_2
    assert m.final_asset_address == FINAL_ETH_ASSET
    assert m.min_amount_out == 333888
    assert m.s_swap_interval == 0
    assert m.s_swap_quantity is None
    assert m.build() == f'=:ETH.ETH:{ETH_ADDR}::::{ETH_ADDR_2}:{FINAL_ETH_ASSET}:333888'
    assert THORMemo.parse_memo(f'=:ETH.ETH:{ETH_ADDR}::::{ETH_ADDR_2}:{FINAL_ETH_ASSET}:333888') == m

    m = THORMemo.swap('ETH.ETH', ETH_ADDR, dex_aggregator_address=ETH_ADDR_2,
                      dex_final_asset_address=FINAL_ETH_ASSET, dex_min_amount_out=333888,
                      affiliate_address='t', affiliate_fee_bp=30, limit=9999999, s_swap_interval=4, s_swap_quantity=7)
    assert m.action == ActionType.SWAP
    assert m.pool == 'ETH.ETH'
    assert m.dest_address == ETH_ADDR
    assert m.refund_address == m.dest_address
    assert m.dex_aggregator_address == ETH_ADDR_2
    assert m.final_asset_address == FINAL_ETH_ASSET
    assert m.min_amount_out == 333888
    assert m.limit == 9999999
    assert m.s_swap_interval == 4
    assert m.s_swap_quantity == 7
    assert m.affiliate_address == 't'
    assert m.affiliate_fee_bp == 30
    assert m.build() == f'=:ETH.ETH:{ETH_ADDR}:9999999/4/7:t:30:{ETH_ADDR_2}:{FINAL_ETH_ASSET}:333888'
    assert THORMemo.parse_memo(f'=:ETH.ETH:{ETH_ADDR}:9999999/4/7:t:30:{ETH_ADDR_2}:{FINAL_ETH_ASSET}:333888') == m

    # custom refund address
    m = THORMemo.swap('THOR.RUNE', THOR_ADDR_1, 9873, 33,
                      s_swap_quantity=AUTO_OPTIMIZED,
                      affiliate_address='dx', affiliate_fee_bp=35, refund_address=THOR_ADDR_2)
    assert m.action == ActionType.SWAP
    assert m.pool == 'THOR.RUNE'
    assert m.dest_address == THOR_ADDR_1
    assert m.refund_address == THOR_ADDR_2
    assert m.limit == 9873
    assert m.s_swap_interval == 33
    assert m.s_swap_quantity == AUTO_OPTIMIZED
    assert m.affiliate_address == 'dx'
    assert m.affiliate_fee_bp == 35
    assert m.build() == f'=:THOR.RUNE:{THOR_ADDR_1}/{THOR_ADDR_2}:9873/33/0:dx:35'
    assert THORMemo.parse_memo(f'=:THOR.RUNE:{THOR_ADDR_1}/{THOR_ADDR_2}:9873/33/0:dx:35') == m


def test_add_liquidity():
    m = THORMemo.add_liquidity('ETH.ETH', ETH_ADDR)
    assert m.action == ActionType.ADD_LIQUIDITY
    assert m.pool == 'ETH.ETH'
    assert m.dest_address == ETH_ADDR
    assert m.build() == f'+:ETH.ETH:{ETH_ADDR}'
    assert THORMemo.parse_memo(f'+:ETH.ETH:{ETH_ADDR}') == m

    m = THORMemo.add_liquidity('ETH.ETH', ETH_ADDR, 't', 100)
    assert m.action == ActionType.ADD_LIQUIDITY
    assert m.pool == 'ETH.ETH'
    assert m.dest_address == ETH_ADDR
    assert m.affiliate_address == 't'
    assert m.affiliate_fee_bp == 100
    assert m.build() == f'+:ETH.ETH:{ETH_ADDR}:t:100'
    assert THORMemo.parse_memo(f'+:ETH.ETH:{ETH_ADDR}:t:100') == m

    m = THORMemo.add_savers('ETH/ETH', 't', 100)
    assert m.action == ActionType.ADD_LIQUIDITY
    assert m.pool == 'ETH/ETH'
    assert m.affiliate_address == 't'
    assert m.affiliate_fee_bp == 100
    assert m.build() == f'+:ETH/ETH::t:100'
    assert THORMemo.parse_memo(f'+:ETH/ETH::t:100') == m

    m = THORMemo.add_liquidity('ETH.ETH', ETH_ADDR, affiliates=[
        Affiliate('t', 30), Affiliate('dx', 40), Affiliate(THOR_ADDR_1, 15)
    ])
    assert m.action == ActionType.ADD_LIQUIDITY
    assert m.pool == 'ETH.ETH'
    assert m.dest_address == ETH_ADDR
    assert m.affiliates == [
        Affiliate('t', 30), Affiliate('dx', 40), Affiliate(THOR_ADDR_1, 15)
    ]


def test_trade_asset():
    m = THORMemo.deposit_trade_account(THOR_ADDR_1)
    assert m.action == ActionType.TRADE_ACC_DEPOSIT
    assert m.dest_address == THOR_ADDR_1
    assert m.build() == f'TRADE+:{THOR_ADDR_1}'
    assert THORMemo.parse_memo(f'TRADE+:{THOR_ADDR_1}') == m

    m = THORMemo.withdraw_trade_account(ETH_ADDR)
    assert m.action == ActionType.TRADE_ACC_WITHDRAW
    assert m.dest_address == ETH_ADDR
    assert m.build() == f'TRADE-:{ETH_ADDR}'
    assert THORMemo.parse_memo(f'TRADE-:{ETH_ADDR}') == m

    m = THORMemo.parse_memo('tRaDe-')
    assert m.action == ActionType.TRADE_ACC_WITHDRAW
    assert m.dest_address == ''


def test_memo_runepool():
    m = THORMemo.runepool_add()
    assert m.action == ActionType.RUNEPOOL_ADD
    assert m.build() == 'POOL+'
    assert THORMemo.parse_memo('POOL+') == m
    assert THORMemo.parse_memo('pool+') == m
    assert THORMemo.parse_memo('Pool+') == m

    m = THORMemo.runepool_withdraw(9955, 't', 23)
    assert m.action == ActionType.RUNEPOOL_WITHDRAW
    assert m.withdraw_portion_bp == 9955
    assert m.affiliate_address == 't'
    assert m.affiliate_fee_bp == 23
    assert m.build() == 'POOL-:9955:t:23'

    m = THORMemo.runepool_withdraw(10000)
    assert m.action == ActionType.RUNEPOOL_WITHDRAW
    assert m.withdraw_portion_bp == 10000
    assert m.affiliate_address == ''
    assert m.affiliate_fee_bp == 0
    assert m.build() == 'POOL-:10000'

    m = THORMemo.runepool_withdraw(3344)
    assert m.action == ActionType.RUNEPOOL_WITHDRAW
    assert m.withdraw_portion_bp == 3344
    assert m.affiliate_address == ''
    assert m.affiliate_fee_bp == 0
    assert m.build() == 'POOL-:3344'


def test_multi_affiliates():
    aff3 = [
        Affiliate('t', 30), Affiliate('dx', 40), Affiliate(THOR_ADDR_1, 15)
    ]
    m = THORMemo.swap('ETH.ETH', ETH_ADDR, affiliates=aff3)
    assert m.action == ActionType.SWAP
    assert m.affiliates == aff3
    assert m.build() == f'=:ETH.ETH:{ETH_ADDR}::t/dx/{THOR_ADDR_1}:30/40/15'

    with pytest.raises(ValueError):
        # 6 is too much
        THORMemo.parse_memo(f'=:ETH.ETH:{ETH_ADDR}::a/b/c/d/e/f:1/2/3/4/5/6')

    with pytest.raises(ValueError):
        # 6 is too much
        THORMemo.swap('ETH.ETH', ETH_ADDR, affiliates=[
            Affiliate('a', 1), Affiliate('b', 2), Affiliate('c', 3),
            Affiliate('d', 4), Affiliate('e', 5), Affiliate('f', 6)
        ])

    # same fee for all
    assert THORMemo.parse_memo(f'=:ETH.ETH:{ETH_ADDR}::a/b/c/d:12').affiliates == [
        Affiliate('a', 12), Affiliate('b', 12), Affiliate('c', 12), Affiliate('d', 12)
    ]

    with pytest.raises(ValueError):
        # mismatch of fee collectors and fee bp number
        THORMemo.parse_memo(f'=:ETH.ETH:{ETH_ADDR}::a/b/c/d:14/15')

    with pytest.raises(ValueError):
        # 1234 is too much (1000 is max)
        THORMemo.parse_memo(f'=:ETH.ETH:{ETH_ADDR}::a/b/c/d:14/0/1234/32')

    assert THORMemo.parse_memo(f'=:ETH.ETH:{ETH_ADDR}::a/b/c/d:14///32').affiliates == [
        Affiliate('a', 14), Affiliate('b', 0), Affiliate('c', 0), Affiliate('d', 32)
    ]

    assert THORMemo.parse_memo(f'=:ETH.ETH:{ETH_ADDR}::a/b/c/d:/3//5').affiliates == [
        Affiliate('a', 0), Affiliate('b', 3), Affiliate('c', 0), Affiliate('d', 5)
    ]

    assert THORMemo.swap('ETH.ETH', ETH_ADDR,
                         affiliates=[('a', 1), ('b', 2), ('c', 3), ('d', 4), ('e', 5)]).affiliates == [
               Affiliate('a', 1), Affiliate('b', 2), Affiliate('c', 3),
               Affiliate('d', 4), Affiliate('e', 5)
           ]



import pytest
from cosmpy.aerial.client import Coin

from xchainpy2_crypto import generate_mnemonic
from xchainpy2_thorchain import THORChainClient
from xchainpy2_utils import NetworkType, CryptoAmount, AssetRUNE, AssetCACAO, Asset, Amount, RUNE_DECIMAL, AssetETH


@pytest.fixture
def client():
    return THORChainClient(phrase=generate_mnemonic())


@pytest.fixture
def stagenet_client():
    return THORChainClient(phrase=generate_mnemonic(), network=NetworkType.STAGENET)


def test_init(client):
    assert client.network == NetworkType.MAINNET
    assert client._decimal == 8
    assert client._gas_asset == AssetRUNE


def test_wallet_index(client):
    for i in range(1, 5):
        client2 = THORChainClient(phrase=client.phrase, wallet_index=i)
        assert client.validate_address(client2.get_address())
        assert client2.get_address() != client.get_address()
        assert client2.get_private_key() != client.get_private_key()
        assert client2.get_public_key() != client.get_public_key()

        assert len(client.get_private_key()) == 64


# noinspection PyTypeChecker
def test_validate_address(client, stagenet_client):
    assert client.validate_address('thor1qd4my7934h2sn5ag5eaqsde39va4ex2asz3yv5')
    assert client.validate_address('thor1505gp5h48zd24uexrfgka70fg8ccedafsnj0e3')
    assert client.validate_address('thor1505gp5h48zd24uexrfgka70fg8ccedafsnj0e3'.upper())

    assert not client.validate_address('')
    assert not client.validate_address('5230')
    assert not client.validate_address(None)
    assert not client.validate_address(0x34434)
    assert not client.validate_address('thor1qd4my7934h2sn5ag5eaqsde39va4ex2asz3yv51')
    assert not client.validate_address('thor1qd4my7934h2sn5ag5eaqsde39va4ex2asz3yv6')
    assert not client.validate_address('sthor1qd4my7934h2sn5ag5eaqsde39va4ex2asz3yv5')
    assert not client.validate_address('tthor1qd4my7934h2sn5ag5eaqsde39va4ex2asz3yv5')
    assert not client.validate_address('cosmos1qd4my7934h2sn5ag5eaqsde39va4ex2asz3yv5')

    assert not stagenet_client.validate_address('thor1qd4my7934h2sn5ag5eaqsde39va4ex2asz3yv5')
    assert stagenet_client.validate_address('sthor16zj3wfurvc6j6sny8s6gt2s3st9yjypqfjegy2')
    assert stagenet_client.validate_address('sthor1g6pnmnyeg48yc3lg796plt0uw50qpp7humfggz')
    assert stagenet_client.validate_address('sthor1g6pnmnyeg48yc3lg796plt0uw50qpp7humfggz'.upper())


def test_address(client, stagenet_client):
    address = client.get_address()
    assert client.validate_address(address)

    assert len(client.get_private_key()) == 64
    assert len(client.get_public_key().public_key_bytes) == 33

    s_addr = stagenet_client.get_address()
    assert s_addr.removeprefix(address)

    assert len(stagenet_client.get_private_key()) == 64


def test_pk_init(client):
    client2 = THORChainClient(private_key=client.get_private_key())
    assert client2.get_address() == client.get_address()
    assert client2.get_private_key() == client.get_private_key()
    assert client2.get_public_key().public_key_bytes == client.get_public_key().public_key_bytes


@pytest.mark.asyncio
async def test_no_keys():
    client = THORChainClient()
    amt = CryptoAmount.zero(AssetRUNE)

    with pytest.raises(Exception):
        # deposit without PK or phrase
        await client.deposit(amt, 'foo')

    with pytest.raises(Exception):
        await client.transfer(amt, client.get_address())


@pytest.mark.parametrize(('asset', 'denom'), [
    (AssetRUNE, 'rune'),
    (Asset.from_string('ETH/ETH'), 'eth/eth'),
    (Asset.from_string('BTC/BTC'), 'btc/btc'),
])
def test_maya_get_denom(client, asset, denom):
    assert client.get_denom(asset) == denom


@pytest.mark.parametrize(('denom', 'asset'), [
    ('atom', None),
    ('cacao', None),
    ('rune', AssetRUNE),
    ('eth/eth', AssetETH.as_synth),
    ('eth~eth', AssetETH.as_trade),
])
def test_maya_parse_denom_to_asset(client, denom, asset):
    assert client.parse_denom_to_asset(denom) == asset


@pytest.mark.parametrize('coin, amount', [
    (Coin(124, 'rune'), CryptoAmount(Amount.from_base(124, RUNE_DECIMAL), AssetRUNE)),
    (Coin(123455, 'eth/eth'), CryptoAmount(Amount.from_base(123455, RUNE_DECIMAL), AssetETH.as_synth))
])
def test_convert_coin_to_amount(client, coin: Coin, amount: CryptoAmount):
    assert client.convert_coin_to_amount(coin) == amount



import sys
from os.path import dirname

sys.path.append(dirname(__file__))


try:
    from .thorchain.v1.common.common_pb2 import Coin as THORCoin, Asset as THORAsset
    from .thorchain.v1.x.thorchain.types.msg_deposit_pb2 import MsgDeposit
    from .thorchain.v1.x.thorchain.types.msg_send_pb2 import MsgSend
    from .cosmos.base.v1beta1.coin_pb2 import Coin as CosmosCoin
except TypeError:
    """
        At the time of writing the code, the structure of these classes is identical in both TC and Maya.
        Therefore, the classes are interchangeable.
        If the classes split up, this approach will no longer work.
    """
    from xchainpy2_mayachain.proto.mayachain.v1.common.common_pb2 import Coin as THORCoin, Asset as THORAsset
    from xchainpy2_mayachain.proto.mayachain.v1.x.mayachain.types.msg_deposit_pb2 import MsgDeposit
    from xchainpy2_mayachain.proto.mayachain.v1.x.mayachain.types.msg_send_pb2 import MsgSend
    from xchainpy2_mayachain.proto.cosmos.base.v1beta1.coin_pb2 import Coin as CosmosCoin



from typing import Optional

# from cosmpy.aerial.client import Coin as CosmosCoin
from cosmpy.aerial.tx import Transaction, SigningCfg
from cosmpy.crypto.keypairs import PublicKey

from xchainpy2_cosmos.utils import convert_address_for_msg
from xchainpy2_utils import NetworkType, CryptoAmount, Amount, RUNE_DECIMAL, Asset, AssetRUNE
from .const import DEPOSIT_GAS_LIMIT_VALUE, DENOM_RUNE_NATIVE
from .proto import THORCoin, THORAsset, MsgSend, MsgDeposit, CosmosCoin


def get_thor_address_prefix(network: NetworkType) -> str:
    """
    Get address prefix based on the network.
    :param network:
    :return: string address prefix
    """
    if network == NetworkType.TESTNET:
        return 'tthor'
    elif network == NetworkType.STAGENET:
        return 'sthor'
    elif network == NetworkType.MAINNET:
        return 'thor'
    else:
        raise ValueError('Invalid network')


def convert_coin_to_crypto_amount(coin: CosmosCoin, decimals=RUNE_DECIMAL) -> CryptoAmount:
    asset = Asset.from_string(f'THOR.{coin.denom.upper()}')
    return CryptoAmount(
        amount=Amount.from_base(coin.amount, decimals),
        asset=asset
    )


def crypto_amount_to_msg_coin(a: CryptoAmount) -> THORCoin:
    asset = THORAsset(
        chain=a.asset.chain,
        symbol=a.asset.full_symbol,
        ticker=a.asset.symbol,
        synth=a.asset.synth,
        trade=a.asset.is_trade,
    )
    return THORCoin(
        asset=asset,
        amount=str(a.amount.as_base.amount), decimals=a.amount.decimals
    )


def build_deposit_tx_unsigned(
        what: CryptoAmount,
        memo: str,
        signer_public_key: PublicKey,
        sequence_num: int,
        prefix: str = 'thor',
        fee=None,
        gas_limit=DEPOSIT_GAS_LIMIT_VALUE,
        second_asset: Optional[CryptoAmount] = None) -> Transaction:
    coins = [
        crypto_amount_to_msg_coin(what)
    ]
    if second_asset is not None:
        coins.append(crypto_amount_to_msg_coin(second_asset))

    tx = Transaction()
    tx.add_message(
        MsgDeposit(coins=coins, memo=memo, signer=convert_address_for_msg(signer_public_key, prefix))
    )

    tx = tx.seal(
        SigningCfg.direct(signer_public_key, sequence_num),
        fee=fee,
        gas_limit=int(gas_limit),
        memo=memo
    )

    return tx


def build_transfer_tx_draft(what: CryptoAmount, denom: str, sender: str, recipient: str,
                            prefix: str) -> Transaction:
    tx = Transaction()

    tx.add_message(
        msg=MsgSend(
            from_address=convert_address_for_msg(sender, prefix),
            to_address=convert_address_for_msg(recipient, prefix),
            amount=[
                CosmosCoin(amount=str(what.amount.internal_amount), denom=denom)
            ],
        )
    )
    return tx


def get_asset_from_denom(denom: str) -> Asset:
    if denom == DENOM_RUNE_NATIVE:
        return AssetRUNE
    else:
        return Asset.from_string_exc(denom.upper())


class NativeTxType:
    DEPOSIT = 'deposit'
    SEND = 'send'
    UNKNOWN = 'unknown'


def get_native_tx_type(raw_json):
    tx_type = raw_json['tx']['body']['messages'][0]['@type']
    if tx_type == "/types.MsgSend":
        return NativeTxType.SEND
    elif tx_type == "/types.MsgDeposit":
        return NativeTxType.DEPOSIT
    else:
        return NativeTxType.UNKNOWN



from gogoproto import gogo_pb2 as _gogo_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class Asset(_message.Message):
    __slots__ = ("chain", "symbol", "ticker", "synth", "trade")
    CHAIN_FIELD_NUMBER: _ClassVar[int]
    SYMBOL_FIELD_NUMBER: _ClassVar[int]
    TICKER_FIELD_NUMBER: _ClassVar[int]
    SYNTH_FIELD_NUMBER: _ClassVar[int]
    TRADE_FIELD_NUMBER: _ClassVar[int]
    chain: str
    symbol: str
    ticker: str
    synth: bool
    trade: bool
    def __init__(self, chain: _Optional[str] = ..., symbol: _Optional[str] = ..., ticker: _Optional[str] = ..., synth: bool = ..., trade: bool = ...) -> None: ...

class Coin(_message.Message):
    __slots__ = ("asset", "amount", "decimals")
    ASSET_FIELD_NUMBER: _ClassVar[int]
    AMOUNT_FIELD_NUMBER: _ClassVar[int]
    DECIMALS_FIELD_NUMBER: _ClassVar[int]
    asset: Asset
    amount: str
    decimals: int
    def __init__(self, asset: _Optional[_Union[Asset, _Mapping]] = ..., amount: _Optional[str] = ..., decimals: _Optional[int] = ...) -> None: ...

class PubKeySet(_message.Message):
    __slots__ = ("secp256k1", "ed25519")
    SECP256K1_FIELD_NUMBER: _ClassVar[int]
    ED25519_FIELD_NUMBER: _ClassVar[int]
    secp256k1: str
    ed25519: str
    def __init__(self, secp256k1: _Optional[str] = ..., ed25519: _Optional[str] = ...) -> None: ...

class Tx(_message.Message):
    __slots__ = ("id", "chain", "from_address", "to_address", "coins", "gas", "memo")
    ID_FIELD_NUMBER: _ClassVar[int]
    CHAIN_FIELD_NUMBER: _ClassVar[int]
    FROM_ADDRESS_FIELD_NUMBER: _ClassVar[int]
    TO_ADDRESS_FIELD_NUMBER: _ClassVar[int]
    COINS_FIELD_NUMBER: _ClassVar[int]
    GAS_FIELD_NUMBER: _ClassVar[int]
    MEMO_FIELD_NUMBER: _ClassVar[int]
    id: str
    chain: str
    from_address: str
    to_address: str
    coins: _containers.RepeatedCompositeFieldContainer[Coin]
    gas: _containers.RepeatedCompositeFieldContainer[Coin]
    memo: str
    def __init__(self, id: _Optional[str] = ..., chain: _Optional[str] = ..., from_address: _Optional[str] = ..., to_address: _Optional[str] = ..., coins: _Optional[_Iterable[_Union[Coin, _Mapping]]] = ..., gas: _Optional[_Iterable[_Union[Coin, _Mapping]]] = ..., memo: _Optional[str] = ...) -> None: ...

class Fee(_message.Message):
    __slots__ = ("coins", "pool_deduct")
    COINS_FIELD_NUMBER: _ClassVar[int]
    POOL_DEDUCT_FIELD_NUMBER: _ClassVar[int]
    coins: _containers.RepeatedCompositeFieldContainer[Coin]
    pool_deduct: str
    def __init__(self, coins: _Optional[_Iterable[_Union[Coin, _Mapping]]] = ..., pool_deduct: _Optional[str] = ...) -> None: ...

class ProtoUint(_message.Message):
    __slots__ = ("value",)
    VALUE_FIELD_NUMBER: _ClassVar[int]
    value: str
    def __init__(self, value: _Optional[str] = ...) -> None: ...



import json

import pytest
from aioresponses import aioresponses

from xchainpy2_client import FeeType, FeeOption
from xchainpy2_thorchain import THORChainClient
from xchainpy2_utils import RUNE_DECIMAL, Amount



def load_json(file_name):
    if not file_name.endswith('.json'):
        file_name = f'{file_name}.json'
    with open(f'./mock/{file_name}') as f:
        return json.load(f)


@pytest.mark.asyncio
async def test_fees_thornode():
    with aioresponses() as m:
        client = THORChainClient()

        m.get(f'https://thornode.ninerealms.com/thorchain/network', payload=load_json('network'))

        fees = await client.get_fees()
        assert fees.type == FeeType.FLAT_FEE
        assert fees.fees[FeeOption.AVERAGE] == fees.fees[FeeOption.FAST] == fees.fees[FeeOption.FASTEST] == \
            Amount.from_base(3120509, RUNE_DECIMAL)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: thorchain/v1/common/common.proto
# Protobuf Python Version: 5.26.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from gogoproto import gogo_pb2 as gogoproto_dot_gogo__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n thorchain/v1/common/common.proto\x12\x06\x63ommon\x1a\x14gogoproto/gogo.proto\"w\n\x05\x41sset\x12\x18\n\x05\x63hain\x18\x01 \x01(\tB\t\xfa\xde\x1f\x05\x43hain\x12\x1a\n\x06symbol\x18\x02 \x01(\tB\n\xfa\xde\x1f\x06Symbol\x12\x1a\n\x06ticker\x18\x03 \x01(\tB\n\xfa\xde\x1f\x06Ticker\x12\r\n\x05synth\x18\x04 \x01(\x08\x12\r\n\x05trade\x18\x05 \x01(\x08\"}\n\x04\x43oin\x12\"\n\x05\x61sset\x18\x01 \x01(\x0b\x32\r.common.AssetB\x04\xc8\xde\x1f\x00\x12?\n\x06\x61mount\x18\x02 \x01(\tB/\xc8\xde\x1f\x00\xda\xde\x1f\'github.com/cosmos/cosmos-sdk/types.Uint\x12\x10\n\x08\x64\x65\x63imals\x18\x03 \x01(\x03\"T\n\tPubKeySet\x12*\n\tsecp256k1\x18\x01 \x01(\tB\x17\xe2\xde\x1f\tSecp256k1\xfa\xde\x1f\x06PubKey\x12\x1b\n\x07\x65\x64\x32\x35\x35\x31\x39\x18\x02 \x01(\tB\n\xfa\xde\x1f\x06PubKey\"\xe0\x01\n\x02Tx\x12\x1a\n\x02id\x18\x01 \x01(\tB\x0e\xe2\xde\x1f\x02ID\xfa\xde\x1f\x04TxID\x12\x18\n\x05\x63hain\x18\x02 \x01(\tB\t\xfa\xde\x1f\x05\x43hain\x12!\n\x0c\x66rom_address\x18\x03 \x01(\tB\x0b\xfa\xde\x1f\x07\x41\x64\x64ress\x12\x1f\n\nto_address\x18\x04 \x01(\tB\x0b\xfa\xde\x1f\x07\x41\x64\x64ress\x12*\n\x05\x63oins\x18\x05 \x03(\x0b\x32\x0c.common.CoinB\r\xc8\xde\x1f\x00\xaa\xdf\x1f\x05\x43oins\x12&\n\x03gas\x18\x06 \x03(\x0b\x32\x0c.common.CoinB\x0b\xc8\xde\x1f\x00\xaa\xdf\x1f\x03Gas\x12\x0c\n\x04memo\x18\x07 \x01(\t\"w\n\x03\x46\x65\x65\x12*\n\x05\x63oins\x18\x01 \x03(\x0b\x32\x0c.common.CoinB\r\xc8\xde\x1f\x00\xaa\xdf\x1f\x05\x43oins\x12\x44\n\x0bpool_deduct\x18\x02 \x01(\tB/\xc8\xde\x1f\x00\xda\xde\x1f\'github.com/cosmos/cosmos-sdk/types.Uint\"Q\n\tProtoUint\x12>\n\x05value\x18\x01 \x01(\tB/\xc8\xde\x1f\x00\xda\xde\x1f\'github.com/cosmos/cosmos-sdk/types.Uint:\x04\x80\xdc \x01\x42\x32Z$gitlab.com/thorchain/thornode/common\xc8\xe1\x1e\x00\xd8\xe1\x1e\x00\x80\xe2\x1e\x00\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'thorchain.v1.common.common_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'Z$gitlab.com/thorchain/thornode/common\310\341\036\000\330\341\036\000\200\342\036\000'
  _globals['_ASSET'].fields_by_name['chain']._loaded_options = None
  _globals['_ASSET'].fields_by_name['chain']._serialized_options = b'\372\336\037\005Chain'
  _globals['_ASSET'].fields_by_name['symbol']._loaded_options = None
  _globals['_ASSET'].fields_by_name['symbol']._serialized_options = b'\372\336\037\006Symbol'
  _globals['_ASSET'].fields_by_name['ticker']._loaded_options = None
  _globals['_ASSET'].fields_by_name['ticker']._serialized_options = b'\372\336\037\006Ticker'
  _globals['_COIN'].fields_by_name['asset']._loaded_options = None
  _globals['_COIN'].fields_by_name['asset']._serialized_options = b'\310\336\037\000'
  _globals['_COIN'].fields_by_name['amount']._loaded_options = None
  _globals['_COIN'].fields_by_name['amount']._serialized_options = b'\310\336\037\000\332\336\037\'github.com/cosmos/cosmos-sdk/types.Uint'
  _globals['_PUBKEYSET'].fields_by_name['secp256k1']._loaded_options = None
  _globals['_PUBKEYSET'].fields_by_name['secp256k1']._serialized_options = b'\342\336\037\tSecp256k1\372\336\037\006PubKey'
  _globals['_PUBKEYSET'].fields_by_name['ed25519']._loaded_options = None
  _globals['_PUBKEYSET'].fields_by_name['ed25519']._serialized_options = b'\372\336\037\006PubKey'
  _globals['_TX'].fields_by_name['id']._loaded_options = None
  _globals['_TX'].fields_by_name['id']._serialized_options = b'\342\336\037\002ID\372\336\037\004TxID'
  _globals['_TX'].fields_by_name['chain']._loaded_options = None
  _globals['_TX'].fields_by_name['chain']._serialized_options = b'\372\336\037\005Chain'
  _globals['_TX'].fields_by_name['from_address']._loaded_options = None
  _globals['_TX'].fields_by_name['from_address']._serialized_options = b'\372\336\037\007Address'
  _globals['_TX'].fields_by_name['to_address']._loaded_options = None
  _globals['_TX'].fields_by_name['to_address']._serialized_options = b'\372\336\037\007Address'
  _globals['_TX'].fields_by_name['coins']._loaded_options = None
  _globals['_TX'].fields_by_name['coins']._serialized_options = b'\310\336\037\000\252\337\037\005Coins'
  _globals['_TX'].fields_by_name['gas']._loaded_options = None
  _globals['_TX'].fields_by_name['gas']._serialized_options = b'\310\336\037\000\252\337\037\003Gas'
  _globals['_FEE'].fields_by_name['coins']._loaded_options = None
  _globals['_FEE'].fields_by_name['coins']._serialized_options = b'\310\336\037\000\252\337\037\005Coins'
  _globals['_FEE'].fields_by_name['pool_deduct']._loaded_options = None
  _globals['_FEE'].fields_by_name['pool_deduct']._serialized_options = b'\310\336\037\000\332\336\037\'github.com/cosmos/cosmos-sdk/types.Uint'
  _globals['_PROTOUINT'].fields_by_name['value']._loaded_options = None
  _globals['_PROTOUINT'].fields_by_name['value']._serialized_options = b'\310\336\037\000\332\336\037\'github.com/cosmos/cosmos-sdk/types.Uint'
  _globals['_PROTOUINT']._loaded_options = None
  _globals['_PROTOUINT']._serialized_options = b'\200\334 \001'
  _globals['_ASSET']._serialized_start=66
  _globals['_ASSET']._serialized_end=185
  _globals['_COIN']._serialized_start=187
  _globals['_COIN']._serialized_end=312
  _globals['_PUBKEYSET']._serialized_start=314
  _globals['_PUBKEYSET']._serialized_end=398
  _globals['_TX']._serialized_start=401
  _globals['_TX']._serialized_end=625
  _globals['_FEE']._serialized_start=627
  _globals['_FEE']._serialized_end=746
  _globals['_PROTOUINT']._serialized_start=748
  _globals['_PROTOUINT']._serialized_end=829
# @@protoc_insertion_point(module_scope)



from cosmpy.aerial.client import Coin

from xchainpy2_thorchain import convert_coin_to_crypto_amount, get_thor_address_prefix, get_asset_from_denom, \
    crypto_amount_to_msg_coin
from xchainpy2_utils import CryptoAmount, Amount, AssetRUNE, RUNE_DECIMAL, Asset, NetworkType, AssetBNB


def test_convert_coin_to_crypto_amount():
    convert = convert_coin_to_crypto_amount
    assert convert(Coin(100000000, 'rune')) == CryptoAmount(Amount.from_asset(1, RUNE_DECIMAL).as_base, AssetRUNE)
    assert convert(Coin(10, 'uatom')) == CryptoAmount(Amount.from_base(10), Asset.from_string('THOR.UATOM'))
    assert convert(Coin(333, 'uatom')) == CryptoAmount.from_base('333', Asset.from_string('THOR.UATOM'))

    a = convert(Coin(7777, 'btc/btc'))
    assert a.amount == Amount.from_base(7777)
    assert a.asset == Asset.from_string('THOR.BTC/BTC')
    assert a.asset.synth


def test_prefix():
    assert get_thor_address_prefix(NetworkType.MAINNET) == 'thor'
    assert get_thor_address_prefix(NetworkType.TESTNET) == 'tthor'
    assert get_thor_address_prefix(NetworkType.STAGENET) == 'sthor'


def test_asset_from_denom():
    assert get_asset_from_denom('rune') == AssetRUNE
    assert get_asset_from_denom('bnb.bnb') == AssetBNB


def test_crypto_amount_to_msg_coin():
    a = crypto_amount_to_msg_coin(CryptoAmount(
        Amount.from_base(100, 8),
        AssetRUNE
    ))

    assert a.decimals == 8
    assert a.amount == "100"
    assert not a.asset.synth
    assert a.asset.chain == 'THOR'
    assert a.asset.ticker == 'RUNE'
    assert a.asset.symbol == 'RUNE'

    a = crypto_amount_to_msg_coin(CryptoAmount(
        Amount.from_base(333, 6),
        Asset.from_string('BTC/BTC')
    ))
    assert a.decimals == 6
    assert a.amount == "333"
    assert a.asset.synth
    assert a.asset.chain == 'BTC'
    assert a.asset.ticker == 'BTC'
    assert a.asset.symbol == 'BTC'

    a = crypto_amount_to_msg_coin(CryptoAmount(
        Amount.from_base(987654321, 18),
        Asset.from_string('ETH.ETH-0xdac17f958d2ee523a2206206994597c13d831ec7')
    ))
    assert a.decimals == 18
    assert a.amount == "987654321"
    assert not a.asset.synth
    assert a.asset.chain == 'ETH'
    assert a.asset.ticker == 'ETH'
    assert a.asset.symbol == 'ETH-0xdac17f958d2ee523a2206206994597c13d831ec7'



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: thorchain/v1/x/thorchain/types/msg_deposit.proto
# Protobuf Python Version: 5.26.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from thorchain.v1.common import common_pb2 as thorchain_dot_v1_dot_common_dot_common__pb2
from gogoproto import gogo_pb2 as gogoproto_dot_gogo__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n0thorchain/v1/x/thorchain/types/msg_deposit.proto\x12\x05types\x1a thorchain/v1/common/common.proto\x1a\x14gogoproto/gogo.proto\"\xae\x01\n\nMsgDeposit\x12O\n\x05\x63oins\x18\x01 \x03(\x0b\x32\x0c.common.CoinB2\xc8\xde\x1f\x00\xaa\xdf\x1f*gitlab.com/thorchain/thornode/common.Coins\x12\x0c\n\x04memo\x18\x02 \x01(\t\x12\x41\n\x06signer\x18\x03 \x01(\x0c\x42\x31\xfa\xde\x1f-github.com/cosmos/cosmos-sdk/types.AccAddressB1Z/gitlab.com/thorchain/thornode/x/thorchain/typesb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'thorchain.v1.x.thorchain.types.msg_deposit_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'Z/gitlab.com/thorchain/thornode/x/thorchain/types'
  _globals['_MSGDEPOSIT'].fields_by_name['coins']._loaded_options = None
  _globals['_MSGDEPOSIT'].fields_by_name['coins']._serialized_options = b'\310\336\037\000\252\337\037*gitlab.com/thorchain/thornode/common.Coins'
  _globals['_MSGDEPOSIT'].fields_by_name['signer']._loaded_options = None
  _globals['_MSGDEPOSIT'].fields_by_name['signer']._serialized_options = b'\372\336\037-github.com/cosmos/cosmos-sdk/types.AccAddress'
  _globals['_MSGDEPOSIT']._serialized_start=116
  _globals['_MSGDEPOSIT']._serialized_end=290
# @@protoc_insertion_point(module_scope)



from typing import NamedTuple, Dict

from xchainpy2_client import ExplorerProvider
from xchainpy2_utils import NetworkType


class NodeURL(NamedTuple):
    node: str
    rpc: str

    @classmethod
    def from_ip_address(cls, ip_address: str, thornode_port: int, rpc_port: int, protocol='http'):
        return cls(
            f'{protocol}://{ip_address}:{thornode_port}',
            f'{protocol}://{ip_address}:{rpc_port}',
        )


class RPCTxResultInner(NamedTuple):
    code: int
    data: str
    log: str
    info: str
    gas_wanted: str
    gas_used: str
    events: list
    codespace: str


class RPCTxResult(NamedTuple):
    hash: str
    height: int
    index: int
    tx_results: RPCTxResultInner
    tx: str



from decimal import Decimal

from xchainpy2_client import ExplorerProvider
from xchainpy2_utils import Asset, Amount, CryptoAmount, AssetRUNE
from xchainpy2_utils.consts import NetworkType, RUNE_DECIMAL
from .models import NodeURL

DEFAULT_GAS_LIMIT_VALUE = 6_000_000
"""Default gas limit for transactions. Refers to Cosmos SDK"""

DEPOSIT_GAS_LIMIT_VALUE = 600_000_000
"""Default gas limit for deposit transactions. Refers to Cosmos SDK"""

MAX_TX_COUNT_PER_PAGE = 100
"""Maximum number of transactions per page."""

MAX_TX_COUNT_PER_FUNCTION_CALL = 500
"""Maximum number of transactions per function call."""

MAX_PAGES_PER_FUNCTION_CALL = 15
"""Maximum number of pages per function call."""

RUNE_DISPLAY_SYMBOL = 'ᚱ'
"""Display symbol for the RUNE token. Raido glyph from the Elder Futhark alphabet."""

RUNE_TICKER = 'RUNE'
"""Ticker symbol for the RUNE token."""

DEFAULT_RUNE_FEE = Amount.from_asset(0.02, RUNE_DECIMAL)
"""Default fee for RUNE transactions. Subject to change."""

DEFAULT_EXPLORER_URL = 'https://runescan.io{path}{network_tag}'
"""Default explorer URL for the THORChain network."""

THOR_BASIS_POINT_MAX = 10_000
"""Maximum basis point value for THORChain."""

THOR_AFFILIATE_BASIS_POINT_MAX = 1_000
"""Affiliate cannot receive more than 10% of the trade value."""


def make_explorer_object(network_tag: str):
    """
    Make an explorer object.

    :param network_tag: Query string to append to the base URL
    :return: ExplorerProvider object
    """
    return ExplorerProvider(
        DEFAULT_EXPLORER_URL.format(network_tag=network_tag, path=''),
        DEFAULT_EXPLORER_URL.format(network_tag=network_tag, path='/address/{address}'),
        DEFAULT_EXPLORER_URL.format(network_tag=network_tag, path='/tx/{tx_id}'),
    )


THOR_EXPLORERS = {
    NetworkType.MAINNET: make_explorer_object(''),
    NetworkType.TESTNET: make_explorer_object('?network=testnet'),
    NetworkType.STAGENET: make_explorer_object('?network=stagenet')
}
"""Explorer objects for the THORChain network."""

ROOT_DERIVATION_PATH = "44'/931'/0'/0/"
"""Root derivation path of crypto-keys for THORChain network."""

ROOT_DERIVATION_PATHS = {
    NetworkType.MAINNET: ROOT_DERIVATION_PATH,
    NetworkType.STAGENET: ROOT_DERIVATION_PATH,
    NetworkType.TESTNET: ROOT_DERIVATION_PATH,
}
"""Root derivation paths for the THORChain network."""

DEFAULT_CHAIN_IDS = {
    NetworkType.MAINNET: 'thorchain-1',  # todo: watch out for hard fork
    NetworkType.STAGENET: 'thorchain-stagenet-2',  # todo: watch out for hard fork
    NetworkType.TESTNET: 'deprecated',
}
"""Default chain IDs for the THORChain network. Subject to change."""

THORNODE_PORT = 1317
"""Default API port for THORNode."""

RPC_PORTS = {
    NetworkType.MAINNET: 27147,
    NetworkType.STAGENET: 26657,
    NetworkType.TESTNET: 26657,
}
"""Cosmos SDK RPC ports for the THORChain network."""


def make_client_urls_from_ip_address(ip_address: str, network=NetworkType.MAINNET, protocol='http'):
    """
    Make client URLs dict from an IP address.

    :param ip_address: THORNode IP address
    :param network: Network type
    :param protocol: HTTP or HTTPS
    :return: dict of client URLs
    """
    rpc_port = RPC_PORTS[network]
    return {
        network: NodeURL.from_ip_address(ip_address, THORNODE_PORT, rpc_port, protocol)
    }


DEFAULT_CLIENT_URLS = {
    NetworkType.MAINNET: NodeURL(
        'https://thornode.ninerealms.com',
        'https://rpc.ninerealms.com'
    ),
    NetworkType.STAGENET: NodeURL(
        'https://stagenet-thornode.ninerealms.com',
        'https://stagenet-rpc.ninerealms.com'
    ),
    NetworkType.TESTNET: NodeURL('deprecated', 'deprecated'),
}
"""Default client URLs for the THORChain network."""

FALLBACK_CLIENT_URLS = {
    NetworkType.MAINNET: [
        NodeURL('https://thornode-v1.ninerealms.com', 'https://rpc-v1.ninerealms.com'),
        NodeURL('https://thornode.thorswap.net/', 'https://rpc.thorswap.net'),
        NodeURL('https://thornode-v0.ninerealms.com', 'https://rpc-v0.ninerealms.com'),
    ],
    NetworkType.STAGENET: [DEFAULT_CLIENT_URLS[NetworkType.STAGENET]],
    NetworkType.TESTNET: []
}
"""Fallback client URLs for the THORChain network."""

AssetRuneBNBTestnet = Asset.from_string('BNB.RUNE-67C')
"""BEP2 RUNE asset on Binance testnet. No longer used."""

AssetRuneBNBMainnet = Asset.from_string('BNB.RUNE-B1A')
"""BEP2 RUNE asset on Binance mainnet. No longer used."""

AssetRuneERC20Mainnet = Asset.from_string('ETH.RUNE-0x3155ba85d5f96b2d030a4966af206230e46849cb')
"""ERC20 RUNE asset on Ethereum mainnet. No longer used."""

# Base "chain" asset for RUNE on ethereum test net.
AssetRuneERC20Testnet = Asset.from_string('ETH.RUNE-0xd601c6A3a36721320573885A8d8420746dA3d7A0')
"""ERC20 RUNE asset on Ethereum testnet. No longer used."""

DENOM_RUNE_NATIVE = 'rune'
"""Denomination for RUNE on THORChain (COSMOS demon)."""

THOR_BLOCK_TIME_SEC = 6.0
"""Typical time in seconds for a block to be produced in THORChain."""

DEFAULT_RUNE_NETWORK_FEE = CryptoAmount(Amount.from_asset(Decimal("0.02"), RUNE_DECIMAL), AssetRUNE)
"""
    Default network fee for RUNE transactions. 
    This may be overridden by the network constants. 
    Please check the network constants for the actual network fee.
"""



import asyncio
from typing import Optional, Union, List

from aiohttp import ClientSession
from bip_utils import Bech32ChecksumError
from cosmpy.aerial.tx import Transaction
from cosmpy.aerial.tx_helpers import SubmittedTx

from xchainpy2_client import RootDerivationPaths, FeeBounds
from xchainpy2_client import XcTx, TxType, Fees, FeeType, TokenTransfer
from xchainpy2_client.fees import single_fee
from xchainpy2_cosmos import CosmosGaiaClient, TxLoadException, TxInternalException
from xchainpy2_cosmos.utils import parse_tx_response_json
from xchainpy2_crypto import decode_address
from xchainpy2_thornode import ApiClient, NetworkApi, TradeAccountApi, TradeAccountResponse
from xchainpy2_utils import Chain, NetworkType, AssetRUNE, RUNE_DECIMAL, CryptoAmount, Amount, remove_0x_prefix, \
    Asset, AssetKind
from .const import NodeURL, DEFAULT_CHAIN_IDS, DEFAULT_CLIENT_URLS, DENOM_RUNE_NATIVE, ROOT_DERIVATION_PATHS, \
    THOR_EXPLORERS, DEFAULT_GAS_LIMIT_VALUE, DEPOSIT_GAS_LIMIT_VALUE, FALLBACK_CLIENT_URLS, DEFAULT_RUNE_FEE, \
    make_client_urls_from_ip_address
from .utils import get_thor_address_prefix, build_deposit_tx_unsigned, build_transfer_tx_draft


class THORChainClient(CosmosGaiaClient):
    @classmethod
    def from_node_ip(cls, ip: str):
        """
        Initialize THORChainClient from node IP address.

        :param ip: IP address of the node
        :return: THORChainClient
        """
        return cls(client_urls=make_client_urls_from_ip_address(ip))

    def __init__(self,
                 network=NetworkType.MAINNET,
                 phrase: Optional[str] = None,
                 private_key: Union[str, bytes, callable, None] = None,
                 fee_bound: Optional[FeeBounds] = None,
                 root_derivation_paths: Optional[RootDerivationPaths] = None,
                 client_urls=DEFAULT_CLIENT_URLS,
                 fallback_client_urls=FALLBACK_CLIENT_URLS,
                 chain_ids=DEFAULT_CHAIN_IDS,
                 explorer_providers=THOR_EXPLORERS,
                 wallet_index=0,
                 ):
        """
        Initialize THORChainClient.

        :param network: Network type. Default is `NetworkType.MAINNET`
        :param phrase: Mnemonic phrase
        :param private_key: Private key (if you want to use a private key instead of a mnemonic phrase)
        :param fee_bound: Fee bound structure. See: FeeBounds
        :param root_derivation_paths: Dictionary of derivation paths for each network type. See: ROOT_DERIVATION_PATHS
        :param client_urls: Dictionary of client urls for each network type. See: DEFAULT_CLIENT_URLS
        :param chain_ids: Dictionary of chain ids for each network type. See: DEFAULT_CHAIN_IDS
        :param explorer_providers: Dictionary of explorer providers for each network type. See: THOR_EXPLORERS
        :param wallet_index: int (wallet index, default 0) We can derive any number of addresses from a single seed
        """
        self.thornode_api_client = ApiClient()

        self.explorers = explorer_providers

        if isinstance(client_urls, NodeURL):
            client_urls = {network: client_urls}

        self._client_urls = client_urls.copy() if client_urls else DEFAULT_CLIENT_URLS.copy()
        self.fallback_client_urls = fallback_client_urls.copy() if fallback_client_urls else None

        self.chain_ids = chain_ids.copy() if chain_ids else DEFAULT_CHAIN_IDS.copy()

        root_derivation_paths = root_derivation_paths.copy() if root_derivation_paths else ROOT_DERIVATION_PATHS.copy()
        super().__init__(
            network, phrase, private_key, fee_bound, root_derivation_paths,
            self._client_urls, self.chain_ids, self.explorers,
            wallet_index
        )
        self._fee_minimum_gas_price = 0

        # Tune for THORChain
        self.chain = Chain.THORChain
        self._prefix = get_thor_address_prefix(network)
        self._gas_asset = AssetRUNE
        self._denom = DENOM_RUNE_NATIVE
        self._decimal = RUNE_DECIMAL
        self._gas_limit = DEFAULT_GAS_LIMIT_VALUE
        self._deposit_gas_limit = DEPOSIT_GAS_LIMIT_VALUE
        self.standard_tx_fee = DEFAULT_RUNE_FEE

        self.set_network(self.network)  # this will set the prefix and client urls for THORNode client
        self._recreate_client()
        self._make_wallet()

    async def close(self):
        """
        Close the client.
        """
        await self.thornode_api_client.rest_client.pool_manager.close()

    def set_network(self, network: NetworkType):
        """
        Set network type for THORChainClient.

        :param network: new network type
        """
        super().set_network(network)
        self._prefix = get_thor_address_prefix(network)
        self.thornode_api_client.configuration.host = self._client_urls[self.network].node

    set_network.__doc__ = CosmosGaiaClient.set_network.__doc__

    @property
    def server_url(self) -> str:
        """
        Get THORNode URL.

        :return: str
        """
        return self._client_urls[self.network].node

    @property
    def rpc_url(self) -> str:
        """
        Get RPC URL for THORChain.

        :return: str
        """
        return self._client_urls[self.network].rpc

    def validate_address(self, address: str) -> bool:
        """
        Validate THORChain address.

        :param address: Address to validate
        :return: bool
        """
        if not super().validate_address(address):
            return False
        try:
            decode_address(address, self._prefix)
            return True
        except (ValueError, Bech32ChecksumError):
            return False

    async def deposit(self,
                      what: Union[CryptoAmount, Amount, int, float],
                      memo: str,
                      second_asset: Optional[CryptoAmount] = None,
                      gas_limit: Optional[int] = None,
                      sequence: int = None,
                      account_number: int = None,
                      check_balance: bool = True,
                      fee=None,
                      return_full_response=False) -> Union[SubmittedTx, str]:
        """
        Send a deposit transaction. MsgDeposit is a special kind of transaction to invoke THORChain protocol's action
        like swap or liquidity addition. Note! It is not on ordinary transfer of tokens. It has no destination
        address.
        For more info see: https://dev.thorchain.org/concepts/sending-transactions.html?highlight=MsgDeposit#thorchain

        :param what: Amount and Asset
        :param second_asset: optional second asset if needed
        :param memo: Memo string (usually a command to the AMM)
        :param gas_limit: if not specified, we'll use the default value
        :param sequence: sequence number. If it is None, it will be fetched automatically
        :param check_balance: Flag to check the balance before sending Tx
        :param fee: string like "0rune", default is 0
        :param account_number: Your account number. If it is none, we will fetch it
        :param return_full_response: when it is not enough to have just tx hash

        :return: submitted tx hash or full response (SubmittedTt object) if requested
        """
        self._throw_if_empty_phrase()

        if isinstance(what, Amount):
            what = CryptoAmount(what, self._gas_asset)
        elif isinstance(what, (int, float)):
            what = CryptoAmount(Amount.automatic(what, self._decimal), self._gas_asset)

        address = self.get_address()

        if check_balance:
            await self.check_balance(address, what)

        if gas_limit is None:
            gas_limit = self._deposit_gas_limit

        if sequence is None or account_number is None:
            account = await self.get_account(address)
            sequence = account.sequence
            account_number = account.number

        public_key = self.get_public_key()

        tx = build_deposit_tx_unsigned(
            what, memo,
            public_key,
            fee=fee or self.get_amount_string(0),
            prefix=self.prefix,
            sequence_num=sequence,
            gas_limit=gas_limit,
            second_asset=second_asset,
        )

        tx.sign(
            self.get_private_key_cosmos(),
            self.get_chain_id(),
            account_number=account_number
        )

        tx.complete()

        result = await asyncio.get_event_loop().run_in_executor(
            None,
            self._client.broadcast_tx,
            tx
        )

        return result if return_full_response else result.tx_hash

    async def fetch_transaction_from_thornode_raw(self, tx_hash: str) -> dict:
        """
        Fetch transaction from THORNode, try to use fallback client if main client is not available
        Url: https://node/thorchain/tx/{tx_hash}
        :param tx_hash: Tx Hash
        :return: Transaction data (raw, unparsed)
        """
        clients = [self._client_urls[self.network]]
        if self.fallback_client_urls:
            clients.extend(self.fallback_client_urls[self.network])

        exc = None
        for client in clients:
            try:
                url = f"{client.node}/thorchain/tx/{tx_hash}"
                j = await self._get_json(url)
                return j
            except Exception as e:
                exc = e
                continue
        if exc:
            raise exc

    async def get_transaction_data_thornode(self, tx_id: str) -> XcTx:
        """
        Fetch transaction data from THORNode (parsed to XcTx instance)
        This function is used when inbound or outbound tx is not of THORChain.
        It is called "getTransactionDataThornode" in xchainjs
        Parsing "observed_tx" object.
        Url: https://node/thorchain/tx/{tx_hash}
        :param tx_id: Tx Hash
        :return: XcTx result
        """
        if not tx_id:
            raise Exception("tx_id is not specified")

        # Remove 0x prefix if exists
        tx_id = remove_0x_prefix(tx_id)

        raw_data = await self.fetch_transaction_from_thornode_raw(tx_id)

        if not raw_data:
            raise TxLoadException(f"Could not fetch transaction data from THORNode {tx_id}")

        error = raw_data.get('error', '')
        if 'desc = internal' in error:
            raise TxInternalException(f"This transaction is internal: {tx_id}, please use cosmos RPC")
        elif error:
            raise TxLoadException(f"Could not fetch transaction data from THORNode {tx_id}. Reason: {error}")

        tx = raw_data['observed_tx']['tx']
        coin = tx['coins'][0]
        sender_asset = Asset.from_string_exc(coin['asset'])
        from_address = tx.get('from_address')
        to_address = tx.get('to_address', 'undefined')
        decimals = coin.get('decimals', self._decimal)
        coin_amount = Amount.from_base(coin['amount'], decimals)
        memo = tx.get('memo', '')
        split_memo = memo.split(':')
        if not split_memo:
            raise TxLoadException('Could not parse memo')
        outbound = split_memo[0].upper() == 'OUT'

        transfers = [
            TokenTransfer(
                from_address, to_address,
                coin_amount, sender_asset, tx_id,
                outbound=outbound
            )
        ]

        height = int(raw_data.get('finalised_height', 0))
        return XcTx(sender_asset, transfers, None, TxType.TRANSFER, tx['id'], height, memo=memo,
                    original=raw_data)

    async def get_transaction_data(self, tx_id: str, address: str = '') -> XcTx:
        """
        Get transaction details by Tx Hash.
        :param tx_id:
        :param address:
        :return: XcTx
        """
        try:
            j = await self.get_transaction_data_cosmos(tx_id)
            return parse_tx_response_json(j, tx_id, address, self._decimal, self._denom, self._gas_asset)
        except TxLoadException:
            return await self.get_transaction_data_thornode(tx_id)

    async def get_fees(self) -> Fees:
        """
        Get THORChain interaction fees from THORNode API.

        :return: Fees object
        """
        network_api = NetworkApi(self.thornode_api_client)
        network_params = await network_api.network()

        fee = network_params.native_tx_fee_rune
        if not fee or not isinstance(fee, str) or not fee.isdigit() or int(fee) < 0:
            raise Exception(f"Invalid fee: {fee}")

        return single_fee(FeeType.FLAT_FEE, Amount.from_base(fee, self._decimal))

    def parse_denom_to_asset(self, denom: str) -> Asset:
        """
        Parse denomination to Asset. RUNE is a special case (just 'rune')

        :param denom: Cosmos coin denomination string
        :return: Asset
        """
        kind = AssetKind.recognize(denom)
        if kind in (AssetKind.SYNTH, AssetKind.TRADE):
            # convert to uppercase and then convert to Asset
            return Asset.from_string(denom.upper())
        elif denom == DENOM_RUNE_NATIVE:
            return AssetRUNE

    def get_denom(self, asset: Asset) -> str:
        """
        Get the denomination of the asset. RUNE is a special case. Other native assets (like synths and trade assets)
        are lowercase.

        :param asset: Asset to get the denomination for
        :return: str
        """
        if asset == AssetRUNE:
            return DENOM_RUNE_NATIVE
        else:
            return str(asset).lower()

    def build_transfer_tx(self, what: CryptoAmount, recipient: str) -> Transaction:
        """
        Build a transfer transaction of THORChain native asset.

        :param what: What to transfer
        :type what: CryptoAmount
        :param recipient: THORChain address of the recipient
        :type recipient: str
        :return: Transaction
        """
        tx = build_transfer_tx_draft(
            what, denom=self.get_denom(what.asset),
            sender=str(self._wallet.address()),
            recipient=recipient,
            prefix=self.prefix,
        )
        return tx

    async def get_trade_asset_balance(self, address: str = '') -> List[CryptoAmount]:
        """
        Get the balance of trade assets of a given address. Trade assets are not listed in the ordinary cosmos/bank.
        This method calls a special THORChain API endpoint.
        See: https://dev.thorchain.org/concepts/trade-accounts.html

        :param address: THORChain address, if not specified, the address of the wallet will be used
        :type address: str
        :return: List of CryptoAmount
        :rtype: List[CryptoAmount]
        """
        if not address:
            address = self.get_address()

        api = TradeAccountApi(self.thornode_api_client)
        result: List[TradeAccountResponse] = await api.trade_account(address)

        return [
            CryptoAmount(
                Amount.from_base(trade_acc.units, decimals=self.decimal),
                Asset.from_string(trade_acc.asset),
            ) for
            trade_acc in result
        ]

    async def get_balance(self, address: str = '', with_trade_accounts=True) -> List[CryptoAmount]:
        """
        Get the native balance of a given address.
        If `with_trade_accounts` is True, the balance will include trade assets, that takes 2 API calls.

        :param address:
        :param with_trade_accounts: Whether include trade account balance or not
        :return: List[CryptoAmount]
        """

        main_balance = await super().get_balance(address)
        if with_trade_accounts:
            main_balance += await self.get_trade_asset_balance(address)
        return main_balance

    @property
    def rest_session(self) -> ClientSession:
        return self.thornode_api_client.rest_client.pool_manager

    async def refresh_chain_id(self):
        """
        Refresh chain ID for the current network.
        """
        rpc = self._client_urls[self.network].rpc
        async with self.rest_session.get(f'{rpc}/status?') as resp:
            data = await resp.json()
            new_chain_id = data['result']['node_info']['network']
            self.chain_ids[self.network] = new_chain_id
            return new_chain_id



"""
See: https://dev.thorchain.org/thorchain-dev/concepts/memos
"""
from dataclasses import dataclass
from enum import Enum
from typing import Union, NamedTuple, Optional, List

from xchainpy2_utils import Chain
from .const import THOR_BASIS_POINT_MAX, RUNE_TICKER, THOR_AFFILIATE_BASIS_POINT_MAX

MAX_AFF_LEVELS = 5


class ActionType(Enum):
    # Standard
    ADD_LIQUIDITY = 'addLiquidity'
    SWAP = 'swap'
    WITHDRAW = 'withdraw'
    DONATE = 'donate'

    # Name service
    THORNAME = 'thorname'

    # Lending
    LOAN_OPEN = 'loan+'
    LOAN_CLOSE = 'loan-'

    # Node operator/bond provider
    BOND = 'bond'
    UNBOND = 'unbond'
    LEAVE = 'leave'

    # Prospective
    LIMIT_ORDER = 'limit_order'

    # RunePool
    RUNEPOOL_ADD = 'pool+'
    RUNEPOOL_WITHDRAW = 'pool-'

    # Outbounds
    REFUND = 'refund'
    OUTBOUND = 'out'

    # Special/dev-centric
    RESERVE = 'reserve'
    NOOP = 'noop'

    TRADE_ACC_DEPOSIT = 'trade+'
    TRADE_ACC_WITHDRAW = 'trade-'

    UNKNOWN = '_unknown_'


MEMO_ACTION_TABLE = {
    "add": ActionType.ADD_LIQUIDITY,
    "+": ActionType.ADD_LIQUIDITY,
    "withdraw": ActionType.WITHDRAW,
    "wd": ActionType.WITHDRAW,
    "-": ActionType.WITHDRAW,
    "swap": ActionType.SWAP,
    "s": ActionType.SWAP,
    "=": ActionType.SWAP,
    "limito": ActionType.LIMIT_ORDER,
    "lo": ActionType.LIMIT_ORDER,
    "out": ActionType.OUTBOUND,
    "donate": ActionType.DONATE,
    "d": ActionType.DONATE,
    "bond": ActionType.BOND,
    "unbond": ActionType.UNBOND,
    "leave": ActionType.LEAVE,
    "reserve": ActionType.RESERVE,
    "refund": ActionType.REFUND,
    "noop": ActionType.NOOP,
    "name": ActionType.THORNAME,
    "n": ActionType.THORNAME,
    "~": ActionType.THORNAME,
    "$+": ActionType.LOAN_OPEN,
    "loan+": ActionType.LOAN_OPEN,
    "$-": ActionType.LOAN_CLOSE,
    "loan-": ActionType.LOAN_CLOSE,
    "trade+": ActionType.TRADE_ACC_DEPOSIT,
    "trade-": ActionType.TRADE_ACC_WITHDRAW,
    "pool+": ActionType.RUNEPOOL_ADD,
    "pool-": ActionType.RUNEPOOL_WITHDRAW,
    # "migrate": TxMigrate,
    # "ragnarok": TxRagnarok,
    # "consolidate": TxConsolidate,
}


def nothing_if_0(x):
    return str(x) if x else ''


AUTO_OPTIMIZED = 0


class Affiliate(NamedTuple):
    address: str
    fee_bp: int


@dataclass
class THORMemo:
    action: ActionType
    asset: str = ''
    dest_address: str = ''
    limit: int = 0
    s_swap_interval: int = 0
    s_swap_quantity: int = 0  # 0 = optimized, 1 = single, >1 = streaming, None = don't care
    affiliates: list[Affiliate] = None
    dex_aggregator_address: str = ''
    final_asset_address: str = ''
    min_amount_out: int = 0
    tx_id: str = ''
    withdraw_portion_bp: int = 0
    pool: str = ''
    node_address: str = ''
    provider_address: str = ''
    amount: int = 0
    no_vault: bool = False
    refund_address: str = ''  # for swaps: dest_addr/refund_addr

    # THORName
    name: str = ''
    owner: str = ''
    chain: str = Chain.THORChain.value
    affiliate_asset: str = ''
    name_expiry: Union[str, int] = ''

    def __str__(self):
        return self.build()

    @property
    def affiliate_address(self) -> str:
        """
        Returns affiliate address/THORName when there is only one affiliate
        Returns '' if no affiliates
        Raises ValueError if multiple affiliates
        :return: Affiliate address/name
        """
        if not self.affiliates:
            return ''
        elif len(self.affiliates) == 1:
            return self.affiliates[0].address
        else:
            raise ValueError(f"Multiple affiliates: {self.affiliates}")

    @property
    def affiliate_fee_bp(self) -> int:
        """
        Returns affiliate fee in basis points (0...1000) when there is only one affiliate
        Returns 0 if no affiliates
        Raises ValueError if multiple affiliates
        :return: Affiliate fee in basis points
        """
        if not self.affiliates:
            return 0
        elif len(self.affiliates) == 1:
            return self.affiliates[0].fee_bp
        else:
            raise ValueError(f"Multiple affiliates: {self.affiliates}")

    @property
    def has_affiliate_part(self):
        return bool(self.affiliates)

    @property
    def is_streaming(self):
        return self.s_swap_quantity > 1

    @property
    def uses_aggregator_out(self):
        return bool(self.dex_aggregator_address)

    @staticmethod
    def parse_dest_address(dest_address: str):
        if '/' in dest_address:
            dest_address, refund_address = dest_address.split('/', maxsplit=1)
        else:
            refund_address = dest_address
        return dest_address.strip(), refund_address.strip()

    @classmethod
    def parse_memo(cls, memo: str, no_raise=False):
        gist, *_comment = memo.split('|', maxsplit=2)  # ignore comments

        components = [it for it in gist.split(':')]

        ith = cls.ith_or_default

        action = ith(components, 0, '').lower()
        tx_type = MEMO_ACTION_TABLE.get(action)

        if tx_type == ActionType.ADD_LIQUIDITY:
            # ADD:POOL:PAIRED_ADDR:AFFILIATE:FEE
            # 0   1    2           3         4
            affiliates = cls._parse_affiliates(
                ith(components, 3, ''),
                ith(components, 4, ''),
            )
            return cls.add_liquidity(
                pool=ith(components, 1, ''),
                paired_address=ith(components, 2, ''),
                affiliates=affiliates,
            )

        elif tx_type == ActionType.SWAP:
            # 0    1     2         3   4         5   6                   7                8
            # SWAP:ASSET:DEST_ADDR:LIM:AFFILIATE:FEE:DEX Aggregator Addr:Final Asset Addr:MinAmountOut
            limit_and_s_swap = ith(components, 3, '')
            limit, s_swap_interval, s_swap_quantity = cls._parse_streaming_params(limit_and_s_swap)
            dest_address, refund_address = cls.parse_dest_address(ith(components, 2, ''))
            affiliates = cls._parse_affiliates(
                ith(components, 4, ''),
                ith(components, 5, ''),
            )
            return cls.swap(
                ith(components, 1),
                dest_address,  # 2
                limit, s_swap_interval, s_swap_quantity,  # 3
                affiliates=affiliates,
                dex_aggregator_address=ith(components, 6, ''),
                dex_final_asset_address=ith(components, 7, ''),
                dex_min_amount_out=ith(components, 8, 0, is_number=True),
                refund_address=refund_address,  # /2
            )

        elif tx_type == ActionType.WITHDRAW:
            # WD:POOL:BASIS_POINTS:ASSET
            # 0  1    2            3
            return cls.withdraw(
                pool=ith(components, 1, ''),
                withdraw_portion_bp=ith(components, 2, THOR_BASIS_POINT_MAX, is_number=True),
                asset=ith(components, 3, ''),
            )

        elif tx_type == ActionType.THORNAME:
            # ~:name:chain:address:?owner:?preferredAsset:?expiry
            # 0 1    2     3       4      5               6
            return cls.thorname_register_or_renew(
                name=ith(components, 1),
                chain=ith(components, 2),
                address=ith(components, 3),
                thor_owner=ith(components, 4, ''),
                preferred_asset=ith(components, 5, ''),
                expiry=ith(components, 6, ''),
            )

        elif tx_type == ActionType.DONATE:
            return cls.donate(
                pool=ith(components, 1, '')
            )

        elif tx_type == ActionType.LOAN_OPEN:
            # LOAN+:BTC.BTC:bc1234567:minBTC:affAddr:affPts:dexAgg:dexTarAddr:DexTargetLimit
            # 0     1       2         3      4       5      6      7          8
            affiliates = cls._parse_affiliates(
                ith(components, 4, ''),
                ith(components, 5, ''),
            )
            return cls.loan_open(
                asset=ith(components, 1),
                dest_address=ith(components, 2),
                limit=ith(components, 3, 0, is_number=True),
                affiliates=affiliates,
                # dex_aggregator_address=ith(components, 6, ''),
                # dex_final_asset_address=ith(components, 7, ''),
                # dex_min_amount_out=ith(components, 8, 0, is_number=True)
            )

        elif tx_type == ActionType.LOAN_CLOSE:
            # "LOAN-:BTC.BTC:bc1234567:minOut"
            #  0     1       2         3

            return cls.loan_close(
                asset=ith(components, 1),
                dest_address=ith(components, 2),
                min_out=ith(components, 3, 0, is_number=True)
            )

        elif tx_type == ActionType.BOND:
            # BOND:NODEADDR:PROVIDER:FEE
            # 0    1        2        3
            return cls.bond(
                node_address=ith(components, 1, ''),
                provider_address=ith(components, 2, ''),
                fee_bp=ith(components, 3, is_number=True),
            )

        elif tx_type == ActionType.UNBOND:
            # UNBOND:NODEADDR:AMOUNT:PROVIDER
            # 0      1        2      3
            return cls.unbond(
                node_address=ith(components, 1, ''),
                amount=ith(components, 2, 0, is_number=True),
                provider_address=ith(components, 3, ''),
            )

        elif tx_type == ActionType.LEAVE:
            # LEAVE:NODEADDR
            # 0     1
            return cls.leave(node_address=ith(components, 1, ''))

        elif tx_type == ActionType.OUTBOUND:
            return cls.outbound(tx_id=ith(components, 1, ''))

        elif tx_type == ActionType.REFUND:
            return cls.refund(tx_id=ith(components, 1, ''))

        elif tx_type == ActionType.RESERVE:
            return cls.reserve()

        elif tx_type == ActionType.NOOP:
            no_vault = ith(components, 1, default='').upper().strip() == 'NOVAULT'
            return cls.noop(no_vault)

        elif tx_type == ActionType.TRADE_ACC_DEPOSIT:
            return cls.deposit_trade_account(dest_address=ith(components, 1, ''))

        elif tx_type == ActionType.TRADE_ACC_WITHDRAW:
            return cls.withdraw_trade_account(dest_address=ith(components, 1, ''))

        elif tx_type == ActionType.RUNEPOOL_ADD:
            return cls.runepool_add()

        elif tx_type == ActionType.RUNEPOOL_WITHDRAW:
            affiliate = ith(components, 2, '')
            affiliate_fee_bp = ith(components, 3, 0)
            return cls.runepool_withdraw(
                bp=ith(components, 1, THOR_BASIS_POINT_MAX, is_number=True),
                affiliates=cls._parse_affiliates(affiliate, affiliate_fee_bp)
            )

        else:
            # todo: limit order, register memo, etc.
            if no_raise:
                return None
            else:
                raise NotImplementedError(f"Not able to parse memo for {action} yet")

    @property
    def _fee_or_empty(self):
        return str(self.affiliate_fee_bp) if self.affiliate_fee_bp is not None else ''

    def build(self):
        if self.action == ActionType.ADD_LIQUIDITY:
            # ADD:POOL:PAIRED_ADDR:AFFILIATE:FEE
            memo = f'+:{self.pool}:{self.dest_address}:{self._affiliate_part}'

        elif self.action == ActionType.SWAP:
            # 0    1     2         3   4         5   6                   7                8
            # SWAP:ASSET:DEST_ADDR:LIM:AFFILIATE:FEE:DEX Aggregator Addr:Final Asset Addr:MinAmountOut
            limit_or_ss = f'{nothing_if_0(self.limit)}'

            # for streaming swaps LIM is like LIM/INTERVAL/QUANTITY
            if self.s_swap_quantity is not None:
                limit_or_ss = f"{limit_or_ss}/{self.s_swap_interval}/{self.s_swap_quantity}"

            dest_addr = self.dest_address
            if self.refund_address and self.refund_address != dest_addr:
                dest_addr += f"/{self.refund_address}"

            memo = (
                f'=:{self.asset}:{dest_addr}:{limit_or_ss}'
                f':{self._affiliate_part}'
                f':{self.dex_aggregator_address}:{self.final_asset_address}:{nothing_if_0(self.min_amount_out)}'
            )

        elif self.action == ActionType.WITHDRAW:
            # -:POOL:BASIS_POINTS:ASSET
            # 0  1    2           3
            memo = f'-:{self.pool}:{nothing_if_0(self.withdraw_portion_bp)}:{self.asset}'

        elif self.action == ActionType.DONATE:
            memo = f'DONATE:{self.pool}'

        elif self.action == ActionType.THORNAME:
            # ~:name:chain:address:?owner:?preferredAsset:?expiry
            # 0 1    2     3       4      5               6
            expiry = self.name_expiry if self.name_expiry is not None else ''
            memo = (
                f'~:{self.name}:{self.chain}:{self.dest_address}:{self.owner}'
                f':{self.affiliate_asset}:{nothing_if_0(expiry)}'
            )

        elif self.action == ActionType.LOAN_OPEN:
            # LOAN+:ASSET:DESTADDR:MINOUT:AFFILIATE:FEE
            memo = (
                f'$+:{self.asset}:{self.dest_address}:{self.limit}'
                f':{self._affiliate_part}'
            )

        elif self.action == ActionType.LOAN_CLOSE:
            # LOAN-:ASSET:DEST_ADDR:MIN_OUT
            memo = f'$-:{self.asset}:{self.dest_address}:{nothing_if_0(self.limit)}'

        elif self.action == ActionType.BOND:
            # # BOND:NODEADDR:PROVIDER:FEE
            if self.provider_address:
                memo = f'BOND:{self.node_address}:{self.provider_address}:{self._fee_or_empty}'
            else:
                memo = f'BOND:{self.node_address}'

        elif self.action == ActionType.UNBOND:
            # UNBOND:NODEADDR:AMOUNT:PROVIDER
            if self.provider_address:
                memo = f'UNBOND:{self.node_address}:{self.amount}:{self.provider_address}'
            else:
                memo = f'UNBOND:{self.node_address}:{self.amount}'

        elif self.action == ActionType.LEAVE:
            memo = f"LEAVE:{self.node_address}"

        elif self.action == ActionType.RESERVE:
            memo = 'RESERVE'

        elif self.action == ActionType.OUTBOUND:
            memo = f'OUT:{self.tx_id}'

        elif self.action == ActionType.REFUND:
            memo = f'REFUND:{self.tx_id}'

        elif self.action == ActionType.NOOP:
            memo = 'NOOP:NOVAULT' if self.no_vault else 'NOOP'

        elif self.action == ActionType.TRADE_ACC_DEPOSIT:
            memo = f'TRADE+:{self.dest_address}'

        elif self.action == ActionType.TRADE_ACC_WITHDRAW:
            memo = f'TRADE-:{self.dest_address}'

        elif self.action == ActionType.RUNEPOOL_ADD:
            memo = 'POOL+'

        elif self.action == ActionType.RUNEPOOL_WITHDRAW:
            memo = f'POOL-:{self.withdraw_portion_bp}:{self._affiliate_part}'

        else:
            raise NotImplementedError(f"Can not build memo for {self.action}")

        return memo.strip().rstrip(':')

    @classmethod
    def add_liquidity(cls, pool: str, paired_address: str = '',
                      affiliate_address: str = '', affiliate_fee_bp: int = 0,
                      affiliates: Optional[List[Affiliate]] = None):
        return cls(
            ActionType.ADD_LIQUIDITY, asset=pool, pool=pool,
            dest_address=paired_address,
            affiliates=cls._form_affiliates(affiliate_address, affiliate_fee_bp, affiliates)
        )

    @classmethod
    def add_savers(cls, pool: str, affiliate_address: str = '', affiliate_fee_bp: int = 0,
                   affiliates: Optional[List[Affiliate]] = None, ):
        assert '/' in pool, "Pool must be synth"
        return cls.add_liquidity(pool, '',
                                 affiliate_address=affiliate_address,
                                 affiliate_fee_bp=affiliate_fee_bp,
                                 affiliates=affiliates)

    @classmethod
    def swap(cls, asset: str, dest_address: str, limit: int = 0, s_swap_interval: int = 0,
             s_swap_quantity: int = None,
             affiliate_address: str = '', affiliate_fee_bp: int = 0,
             affiliates: Optional[List[Affiliate]] = None,
             dex_aggregator_address: str = '', dex_final_asset_address: str = '', dex_min_amount_out: int = 0,
             refund_address: str = ''):
        return cls(
            ActionType.SWAP,
            asset, dest_address, limit,
            s_swap_interval, s_swap_quantity,
            pool=asset,
            affiliates=cls._form_affiliates(affiliate_address, affiliate_fee_bp, affiliates),
            dex_aggregator_address=dex_aggregator_address,
            final_asset_address=dex_final_asset_address,
            min_amount_out=dex_min_amount_out,
            refund_address=refund_address or dest_address,
        )

    @classmethod
    def withdraw(cls, pool: str, withdraw_portion_bp=THOR_BASIS_POINT_MAX, asset: str = ''):
        # WD:POOL:BASIS_POINTS:ASSET
        # 0  1    2           3
        return cls(
            ActionType.WITHDRAW, pool=pool, withdraw_portion_bp=withdraw_portion_bp, asset=asset
        )

    @classmethod
    def withdraw_savers(cls, pool: str, withdraw_portion_bp=THOR_BASIS_POINT_MAX):
        assert '/' in pool, "Pool must be synth"
        return cls.withdraw(pool, withdraw_portion_bp)

    @classmethod
    def withdraw_symmetric(cls, pool: str, withdraw_portion_bp=THOR_BASIS_POINT_MAX):
        return cls.withdraw(pool, withdraw_portion_bp)

    @classmethod
    def withdraw_rune(cls, pool: str, withdraw_portion_bp=THOR_BASIS_POINT_MAX):
        return cls.withdraw(pool, withdraw_portion_bp, asset=RUNE_TICKER)

    @classmethod
    def withdraw_asset(cls, pool: str, withdraw_portion_bp=THOR_BASIS_POINT_MAX):
        return cls.withdraw(pool, withdraw_portion_bp, asset=pool)

    @classmethod
    def donate(cls, pool: str):
        return cls(
            ActionType.DONATE,
            pool=pool, asset=pool
        )

    @classmethod
    def thorname_register_or_renew(cls, name: str, chain: str, address: str, thor_owner: str = '',
                                   preferred_asset: str = '',
                                   expiry=''):
        return cls(
            ActionType.THORNAME,
            name=name,
            chain=chain,
            dest_address=address,
            owner=thor_owner,
            name_expiry=expiry,
            affiliate_asset=preferred_asset,
        )

    @classmethod
    def loan_open(cls, asset: str, dest_address: str, limit: int = 0,
                  affiliate_address: str = '', affiliate_fee_bp: int = 0,
                  affiliates: Optional[List[Affiliate]] = None
                  ):
        # LOAN+:BTC.BTC:bc1234567:minBTC:affAddr:affPts:dexAgg:dexTarAddr:DexTargetLimit
        # 0     1       2         3      4       5      6      7          8
        affiliates = affiliates or ([Affiliate(affiliate_address, affiliate_fee_bp)] if affiliate_address else [])
        return cls(
            ActionType.LOAN_OPEN,
            asset=asset,
            dest_address=dest_address,
            limit=limit,
            affiliates=affiliates,
            pool=asset,
        )

    @classmethod
    def loan_close(cls, asset: str, dest_address: str, min_out: int = 0):
        # "LOAN-:BTC.BTC:bc123456:minOut"
        #  0     1       2         3

        return cls(
            ActionType.LOAN_CLOSE,
            asset=asset, pool=asset,
            dest_address=dest_address, limit=min_out
        )

    @classmethod
    def bond(cls, node_address: str, provider_address: str = '', fee_bp: int = None):
        # BOND:NODEADDR:PROVIDER:FEE
        # 0    1        2        3
        return cls(
            ActionType.BOND,
            node_address=node_address,
            provider_address=provider_address,
            affiliates=[Affiliate(node_address, fee_bp)] if fee_bp else []
        )

    @classmethod
    def unbond(cls, node_address: str, amount: int, provider_address: str = ''):
        # UNBOND:NODEADDR:AMOUNT:PROVIDER
        return cls(
            ActionType.UNBOND,
            node_address=node_address,
            provider_address=provider_address,
            amount=amount,
        )

    @classmethod
    def leave(cls, node_address: str):
        return cls(ActionType.LEAVE, node_address=node_address)

    @classmethod
    def limit_order(cls):
        raise NotImplementedError

    @classmethod
    def refund(cls, tx_id: str):
        return cls(ActionType.REFUND, tx_id=tx_id)

    @classmethod
    def outbound(cls, tx_id: str):
        return cls(ActionType.OUTBOUND, tx_id=tx_id)

    @classmethod
    def reserve(cls):
        # todo
        return cls(ActionType.RESERVE)

    @classmethod
    def noop(cls, no_vault=False):
        return cls(ActionType.NOOP, no_vault=no_vault)

    @classmethod
    def deposit_trade_account(cls, dest_address: str):
        return cls(
            ActionType.TRADE_ACC_DEPOSIT,
            dest_address=dest_address
        )

    @classmethod
    def withdraw_trade_account(cls, dest_address: str):
        return cls(
            ActionType.TRADE_ACC_WITHDRAW,
            dest_address=dest_address
        )

    @classmethod
    def runepool_add(cls):
        return cls(ActionType.RUNEPOOL_ADD)

    @classmethod
    def runepool_withdraw(cls, bp: int,
                          affiliate_address: str = '', affiliate_fee_bp: int = 0,
                          affiliates: Optional[List[Affiliate]] = None):
        return cls(
            ActionType.RUNEPOOL_WITHDRAW,
            withdraw_portion_bp=bp,
            affiliates=cls._form_affiliates(affiliate_address, affiliate_fee_bp, affiliates),
        )

    # Utils:

    @classmethod
    def ith_or_default(cls, a, index, default=None, is_number=False) -> Union[str, int, float]:
        if 0 <= index < len(a):
            try:
                r = a[index].strip()
                if r == '':
                    return default
                return cls._int_read(r) if is_number else r
            except ValueError:
                return default
        else:
            return default

    @classmethod
    def _parse_streaming_params(cls, ss: str):
        s_swap_components = ss.split('/')

        ith = cls.ith_or_default
        limit = ith(s_swap_components, 0, 0, is_number=True)
        s_swap_interval = ith(s_swap_components, 1, 0, is_number=True)
        # 0 = optimized, 1 = single, >1 = streaming, None = don't care
        s_swap_quantity = ith(s_swap_components, 2, is_number=True)
        return limit, s_swap_interval, s_swap_quantity

    @classmethod
    def _parse_affiliates(cls, names_part, fees_part):
        """
        Parses affiliate names and fees.
        :param names_part: A string like "10/20/0/40"
        :param fees_part: A string like "1/thor1/thor2/xyz"
        :return:
        """
        names = names_part.split('/')
        fees = fees_part.split('/') if fees_part.strip() else [0]
        n_fees, n_names = len(fees), len(names)
        if n_names != len(fees) and n_fees != 1:
            raise ValueError(f"Affiliates and fees mismatch: {names_part} vs {fees_part}")

        if n_fees == 1:
            # if there is only one fee, apply it to all affiliates
            fees = [int(fees[0])] * n_names
            n_fees = n_names

        fees = [(int(fee) if fee else 0) for fee in fees]
        for fee in fees:
            cls._guard_affiliate_bp(fee)

        if n_names > MAX_AFF_LEVELS or n_fees > MAX_AFF_LEVELS:
            raise ValueError(f"Too many affiliates: {names_part}:{fees_part}. Max {MAX_AFF_LEVELS}")

        return [Affiliate(name.strip(), fee) for name, fee in zip(names, fees) if name.strip()]

    @property
    def _affiliate_part(self):
        if not self.affiliates:
            return ':'
        if len(self.affiliates) > MAX_AFF_LEVELS:
            raise ValueError(f"Too many affiliates: {self.affiliates}. {MAX_AFF_LEVELS} max.")

        name_part = '/'.join(af.address.strip() for af in self.affiliates)

        fee_equal = all(af.fee_bp == self.affiliates[0].fee_bp for af in self.affiliates)
        fee_part = self.affiliates[0].fee_bp if fee_equal else '/'.join(str(af.fee_bp) for af in self.affiliates)
        return f'{name_part}:{fee_part}'

    @classmethod
    def _int_read(cls, x):
        x = str(x).lower()
        if 'e' in x:
            # e.g.: 232323e5
            return int(float(x))
        else:
            # just int
            return int(x)

    @classmethod
    def _guard_affiliate_bp(cls, affiliate_fee_bp):
        if affiliate_fee_bp < 0 or affiliate_fee_bp > THOR_AFFILIATE_BASIS_POINT_MAX:
            raise ValueError(
                f"Invalid affiliate fee: {affiliate_fee_bp}, must be in [0, {THOR_AFFILIATE_BASIS_POINT_MAX}]")

    @classmethod
    def _form_affiliates(cls,
                         affiliate_address: str = '', affiliate_fee_bp: int = 0,
                         affiliates: Optional[List[Affiliate]] = None) -> List[Affiliate]:
        if affiliates and (affiliate_address or affiliate_fee_bp):
            raise ValueError("Can not have both affiliates and affiliate_address/fee")

        if affiliates:
            if len(affiliates) > MAX_AFF_LEVELS:
                raise ValueError(f"Too many affiliates, max {MAX_AFF_LEVELS}")

            results = []
            for aff in affiliates:
                cls._guard_affiliate_bp(int(aff[1]))
                # we use indices to allow not only NamedTuple but also normal tuples
                results.append(Affiliate(aff[0].strip(), int(aff[1])))
            return results
        else:
            affiliate_fee_bp = int(affiliate_fee_bp)
            cls._guard_affiliate_bp(affiliate_fee_bp)
            return [Affiliate(affiliate_address, affiliate_fee_bp)] if affiliate_address else []



from thorchain.v1.common import common_pb2 as _common_pb2
from gogoproto import gogo_pb2 as _gogo_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class MsgDeposit(_message.Message):
    __slots__ = ("coins", "memo", "signer")
    COINS_FIELD_NUMBER: _ClassVar[int]
    MEMO_FIELD_NUMBER: _ClassVar[int]
    SIGNER_FIELD_NUMBER: _ClassVar[int]
    coins: _containers.RepeatedCompositeFieldContainer[_common_pb2.Coin]
    memo: str
    signer: bytes
    def __init__(self, coins: _Optional[_Iterable[_Union[_common_pb2.Coin, _Mapping]]] = ..., memo: _Optional[str] = ..., signer: _Optional[bytes] = ...) -> None: ...
